# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import logging
import ssl
import sys

import kubernetes_asyncio as kubernetes

from aiohttp import web
from http import HTTPStatus

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class Validator(object):
    def __init__(self):
        self._core_api = kubernetes.client.CoreV1Api()
        self._app = web.Application()
        self._app.add_routes([
            web.get('/healthz', self._handle_healthz),
            web.post('/validate', self._handle_validate),
        ])

    def get_app(self):
        return self._app

    def run(self, host, port, ssl_context=None):
        web.run_app(self.get_app(), host=host, port=port,
                    ssl_context=ssl_context)

    async def _handle_healthz(self, request):
        # Health check.
        return web.Response()

    async def _handle_validate(self, request):
        request_json = await request.json()
        if request_json["request"]["operation"] == "CREATE":
            response = await self._validate_create(request_json["request"])
        elif request_json["request"]["operation"] == "UPDATE":
            response = await self._validate_update(request_json["request"])
        else:
            response = {"allowed": True}
        LOG.info("%s %s/%s: %s",
                 request_json["request"]["operation"],
                 request_json["request"]["namespace"],
                 request_json["request"].get("name", "<none>"),
                 response)
        response["uid"] = request_json["request"]["uid"]
        return web.json_response({
            "apiVersion": "admission.k8s.io/v1",
            "kind": "AdmissionReview",
            "response": response,
        })

    async def _validate_create(self, request):
        job = request["object"]
        namespace = request["namespace"]
        template = {
            "metadata": {"name": "spec.template", "namespace": namespace},
            "template": job["spec"]["template"],
        }
        try:
            await self._core_api.create_namespaced_pod_template(
                namespace, template, dry_run="All")
        except kubernetes.client.rest.ApiException as exc:
            return {
               "allowed": False,
               "status": {
                   "code": HTTPStatus.UNPROCESSABLE_ENTITY,
                   "reason": "Invalid",
                   "message": json.loads(exc.body).get("message"),
               }
            }
        # If maxReplicas is provided, it should be >= minReplicas
        if job["spec"].get("maxReplicas", sys.maxsize) < \
                job["spec"].get("minReplicas", 0):
            return {
               "allowed": False,
               "status": {
                   "code": HTTPStatus.UNPROCESSABLE_ENTITY,
                   "reason": "Invalid",
                   "message": ("spec.maxReplicas must be greater "
                               "than or equal to spec.minReplicas")
               }
            }
        return {"allowed": True}

    async def _validate_update(self, request):
        if request["object"]["spec"] != request["oldObject"]["spec"]:
            return {
                "allowed": False,
                "status": {
                    "code": HTTPStatus.UNPROCESSABLE_ENTITY,
                    "reason": "Forbidden",
                    "message": "updates to job spec are forbidden",
                }
            }
        return {"allowed": True}


if __name__ == "__main__":
    logging.basicConfig()
    kubernetes.config.load_incluster_config()

    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="0.0.0.0")
    parser.add_argument("--port", type=int, default=8080)
    parser.add_argument("--tls-crt", type=str)
    parser.add_argument("--tls-key", type=str)
    args = parser.parse_args()

    if args.tls_crt and args.tls_key:
        ssl_context = ssl.SSLContext()
        ssl_context.load_cert_chain(args.tls_crt, args.tls_key)
    else:
        ssl_context = None

    validator = Validator()
    validator.run(args.host, args.port, ssl_context=ssl_context)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import time
import random
import copy

from collections import Counter
from datetime import datetime, timedelta
from adaptdl.goodput import GoodputFunction, PerfParams, GradParams
from adaptdl_sched.policy.pollux import PolluxPolicy
from adaptdl_sched.policy.speedup import SpeedupFunction
from adaptdl_sched.policy.utils import JobInfo, NodeInfo


@pytest.mark.parametrize("num_nodes", [1, 2, 4, 8])
def test_optimize(num_nodes, total_devices=16):
    # Globals
    N_JOBS = 10
    JOBS = list(range(N_JOBS))
    random.shuffle(JOBS)

    PREEMPTIBLE_IDXS = JOBS[:len(JOBS)//2]
    NON_PREEMPTIBLE_IDXS = JOBS[len(JOBS)//2:]

    assert total_devices % num_nodes == 0
    num_devices = total_devices // num_nodes
    print(f"{num_nodes}x{num_devices} nodes:")
    # Make up a realistic speedup function.
    perf_params = PerfParams(0.121, 0.00568, 0.0236, 0.00634,
                             0.0118, 0.00317, 1.14)
    grad_params = GradParams(sqr=0.00136, var=0.000502)
    goodput_fn = GoodputFunction(perf_params, grad_params, 128)
    speedup_fn = SpeedupFunction(goodput_fn, max_batch_size=1280,
                                 atomic_bsz_range=(64, 256))
    now = datetime.now()
    # Add a node template.
    policy = PolluxPolicy()
    job_resources = {"nvidia.com/gpu": 1, "pods": 1}
    # Add a few nodes.
    node_resources = {"nvidia.com/gpu": num_devices, "pods": 32}
    nodes = {i: NodeInfo(node_resources, preemptible=False)
             for i in range(num_nodes)}
    node_template = NodeInfo(node_resources, preemptible=True)

    # Empty allocations
    prev_allocs = {i: [] for i in JOBS}
    for cycle in range(3):
        # Start allocation cycle
        jobs = {}
        for i in PREEMPTIBLE_IDXS:
            creation_timestamp = now + timedelta(minutes=i),
            jobs[i] = JobInfo(job_resources, speedup_fn, creation_timestamp,
                              min_replicas=0, max_replicas=8)
        for i in NON_PREEMPTIBLE_IDXS:
            creation_timestamp = now + timedelta(minutes=i),
            jobs[i] = JobInfo(job_resources, speedup_fn, creation_timestamp,
                              min_replicas=2, max_replicas=4,
                              preemptible=False)
        start = time.time()
        assert len(jobs) > 0
        allocations, desired_nodes = \
            policy.optimize(jobs, nodes, prev_allocs, node_template)
        duration = time.time() - start
        print(f"optimize {cycle + 1}x ({duration}s sec)")
        node_count = Counter()
        for job_key, placement in allocations.items():
            assert len(placement) <= jobs[job_key].max_replicas
            if placement:
                assert len(placement) >= jobs[job_key].min_replicas
            for node_key in placement:
                node_count[node_key] += 1
        for node_key, count in node_count.items():
            assert count <= nodes[node_key].resources["nvidia.com/gpu"]
            assert count <= nodes[node_key].resources["pods"]

        # Check if we are maintaining allocations for non-preemptible jobs
        for i in NON_PREEMPTIBLE_IDXS:
            if (i in allocations) and prev_allocs[i]:
                assert allocations[i] == prev_allocs[i]

        prev_allocs = copy.deepcopy(allocations)
        # Remove one random job
        remove = random.sample(allocations.keys(), 1)[0]
        if remove in NON_PREEMPTIBLE_IDXS:
            NON_PREEMPTIBLE_IDXS.remove(remove)
            print(f"Deleting non-preemptible job {remove}")
        else:
            PREEMPTIBLE_IDXS.remove(remove)
            print(f"Deleting preemptible job {remove}")
        prev_allocs.pop(remove)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import collections
import copy
import jsonpatch
import kubernetes_asyncio as kubernetes
import logging

import adaptdl_sched.k8s_templates as templates
import adaptdl_sched.config as config

from datetime import datetime, timezone
from prometheus_client import Counter, Summary

from adaptdl_sched.resources import set_default_resources
from adaptdl_sched.utils import patch_job_status

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)

# Prometheus metrics
METRICS_KWARGS = dict(namespace="adaptdl", subsystem="sched")
JOB_SUBMISSION_COUNT = Counter(
    "job_submission_count", "Number of submitted jobs", **METRICS_KWARGS)
JOB_COMPLETION_TIME = Summary(
    "job_completion_time", "Duration of completed jobs",
    labelnames=["status"], **METRICS_KWARGS)


class AdaptDLController(object):
    """
    The main controller responsible for the overall AdaptDLJob lifecycle.
    Essentially, it keeps a queue of AdaptDLJobs whose states may need to be
    synchronized. It watches for events such as pod status changes and
    allocation changes and enqueues any potentially affects AdaptDLJobs. A
    worker coroutine is responsible for processing AdaptDLJobs from the queue
    and guarantees that a single AdaptDLJob is never processed concurrently.
    """

    def __init__(self):
        self._core_api = kubernetes.client.CoreV1Api()
        self._objs_api = kubernetes.client.CustomObjectsApi()
        self._custom_resource = ("adaptdl.petuum.com", "v1",
                                 "", "adaptdljobs")
        self._queue = asyncio.Queue()

    async def run(self):
        # Create service if it doesn't already exist.
        # FIXME: initialize allocations
        await asyncio.gather(
            self._watch_jobs(),
            self._watch_pods(),
            self._sync_worker()
        )

    async def _watch_jobs(self):
        # Watch for changes to AdaptDLJobs and enqueue them to be synced.
        # Perform a full reconcile after every timeout.
        async with kubernetes.watch.Watch() as watch:
            while True:
                async for event in watch.stream(
                        self._objs_api.list_namespaced_custom_object,
                        *self._custom_resource, timeout_seconds=60):
                    job_name = event["object"]["metadata"]["name"]
                    namespace = event["object"]["metadata"]["namespace"]
                    await self._queue.put((namespace, job_name))

    async def _watch_pods(self):
        # Watch for changes to pods and enqueue their AdaptDLJobs to be synced.
        # Perform a full reconcile after every timeout.
        async with kubernetes.watch.Watch() as watch:
            while True:
                async for event in watch.stream(
                        self._core_api.list_namespaced_pod, "",
                        label_selector="adaptdl/job", timeout_seconds=60):
                    pod = event["object"]
                    job_name = pod.metadata.labels["adaptdl/job"]
                    namespace = pod.metadata.namespace
                    await self._queue.put((namespace, job_name))

    async def _sync_worker(self):
        while True:
            (namespace, name) = await self._queue.get()
            await self._sync_job(namespace, name)
            self._queue.task_done()

    async def _sync_job(self, namespace, job_name):
        current_ts = datetime.now(timezone.utc)
        job, pods = await self._get_job_and_pods(namespace, job_name)
        if job is not None:
            job = await self._validate_pods(job, pods, current_ts)
        if job is None:  # Not Found, presumably was deleted.
            await self._delete_pods(pods)
            return
        # Use ChainMap to record updates to the job status fields.
        job["status"] = collections.ChainMap({}, job.get("status", {}))
        # Get the current phase of the job, None if no phase was set.
        allocation = job["status"].get("allocation", [])
        phase = job["status"].setdefault("phase", "Pending")
        replicas = job["status"].get("replicas", 0)
        preemptible = job["spec"].get("preemptible", True)
        if (completion_status := self._detect_completion(pods, preemptible)):
            # Job is already completed.
            job["status"].update(completion_status)
            job["status"].setdefault("completionTimestamp", current_ts)
            job["status"]["allocation"] = allocation = []
            await self._delete_pods(  # Keep failed pods for debug purposes.
                [pod for pod in pods if pod.status.phase != "Failed"])
        elif phase == "Pending":
            if allocation and not pods:
                # Start the next group of pods.
                job["status"]["phase"] = "Starting"
        elif phase == "Starting":
            # FIXME: In case if a pod experiences indefinite ImagePullBackOff,
            # the job can get stuck in the Starting phase
            if (self._count_scheduled_pods(pods) != replicas and
                    self._detect_restart(pods, allocation)) or not allocation:
                # Allocator changed allocation based on new information about
                # resource availability
                job["status"]["phase"] = "Stopping"
            elif allocation and not pods:
                # Start the next group of pods.
                job["status"]["group"] = job["status"].get("group", -1) + 1
                try:
                    new_pods = []
                    for rank in range(len(allocation)):
                        pod = await self._create_pod(
                            job["metadata"],
                            job["spec"]["template"], allocation,
                            job["status"]["group"], rank)
                        new_pods.append(pod)
                except kubernetes.client.rest.ApiException as e:
                    LOG.warning(f"Failed to create pod for {job_name}. "
                                "Setting job status to failed")
                    job["status"]["phase"] = "Failed"
                    job["status"]["reason"] = "PodCreationError"
                    job["status"]["message"] = str(e)
                    await self._delete_pods(new_pods)
            elif len(pods) != replicas:
                # Controller restarted before we can spawn all pods
                job["status"]["phase"] = "Stopping"
            elif self._count_ready_pods(pods) == replicas:
                # all pods are running
                job["status"]["phase"] = "Running"
        elif phase == "Running":
            if self._detect_restart(pods, allocation) or \
                    not pods:
                # 1. Reallocation OR 2. the controller restarted before
                # we can update phase to Stopping
                job["status"]["phase"] = "Stopping"
        elif phase == "Stopping":
            if pods:
                await self._delete_pods(pods)
            else:
                # all pods successfully deleted
                job["status"]["phase"] = "Pending"
        # Set replicas and ready replicas.
        if allocation:
            job["status"]["replicas"] = len(allocation)
            job["status"]["readyReplicas"] = self._count_ready_pods(pods)
        else:
            job["status"]["allocation"] = None
            job["status"]["replicas"] = None
            job["status"]["readyReplicas"] = None
        # Apply changes to AdaptDLJob status.
        patch = {"status": {k: v for k, v in job["status"].maps[0].items()
                            if v != job["status"].maps[1].get(k)}}
        if patch["status"]:
            LOG.info("Patch AdaptDLJob %s: %s", job_name, patch)
            await patch_job_status(self._objs_api, namespace, job_name, patch)

    def _count_ready_pods(self, pods):
        count = 0
        for pod in pods:
            if pod.status.container_statuses and \
                    all(stat.ready for stat in pod.status.container_statuses):
                count += 1
        return count

    def _count_scheduled_pods(self, pods):
        count = 0
        for pod in pods:
            if pod.status.conditions and \
                    any(cond.type == "PodScheduled" and cond.status == "True"
                        for cond in pod.status.conditions):
                count += 1
        return count

    async def _get_job_and_pods(self, namespace, name):
        # Find all pods owned by this AdaptDLJob.
        pods = await self._core_api.list_namespaced_pod(
            namespace, label_selector=f"adaptdl/job={name}")
        pods = pods.items
        # Read a fresh copy of the AdaptDLJob.
        try:
            job = await self._objs_api.get_namespaced_custom_object(
                "adaptdl.petuum.com", "v1", namespace, "adaptdljobs", name)
        except kubernetes.client.rest.ApiException as exc:
            if exc.status == 404:
                return None, pods
            raise  # Unexpected error.
        return job, pods

    async def _validate_pods(self, job, pods, current_ts):
        namespace = job["metadata"]["namespace"]
        name = job["metadata"]["name"]
        patch_status = {}
        # Validate pods for job.
        group_list = []
        replicas_list = []
        rank_list = []
        for pod in pods:
            # Check the expected annotations exist.
            try:
                group = int(pod.metadata.annotations["adaptdl/group"])
                replicas = int(pod.metadata.annotations["adaptdl/replicas"])
                rank = int(pod.metadata.annotations["adaptdl/rank"])
                node = pod.metadata.annotations["adaptdl/node"]
            except (KeyError, ValueError):
                patch_status["phase"] = "Failed"
                patch_status["reason"] = "Invalid"
                patch_status["message"] = \
                    f"invalid annotations for pod {pod.metadata.name}"
                break
            # Check the pod is running on the correct node.
            if pod.spec.node_name and pod.spec.node_name != node:
                patch_status["phase"] = "Failed"
                patch_status["reason"] = "Invalid"
                patch_status["message"] = \
                    f"incorrect node for pod {pod.metadata.name}"
                break
            group_list.append(group)
            replicas_list.append(replicas)
            rank_list.append(rank)
        else:
            # Check all pods have the same group and replicas values, and no
            # pod has a rank which is out of range.
            if len(set(group_list)) > 1 or len(set(replicas_list)) > 1 or \
                    any(rank >= replicas_list[0] for rank in rank_list):
                patch_status["phase"] = "Failed"
                patch_status["reason"] = "Invalid"
                patch_status["message"] = "inconsistent pods in group"
        if patch_status:
            return await patch_job_status(self._objs_api, namespace, name,
                                          {"status": patch_status})
        return job

    def _detect_completion(self, pods, preemptible):
        if not pods:
            return {}

        # Check if all pods succeeded.
        for pod in pods:
            replicas = int(pod.metadata.annotations["adaptdl/replicas"])
            if len(pods) != replicas or pod.status.phase != "Succeeded":
                break
        else:
            return {"phase": "Succeeded"}

        # Check for fatal failures in a list of pods. Non-fatal failures can be
        # due to evictions or temporary resource unavailability.
        def any143(pod):
            if not pod.status.container_statuses:
                return False
            for status in pod.status.container_statuses:
                if (status.state.terminated and
                        status.state.terminated.exit_code == 143):
                    return True
            return False

        for pod in pods:
            if pod.status.phase == "Unknown":
                # This can happen if there's something wrong with the node
                # or kubelet assigned to this pod.
                LOG.warning("Unknown status for pod %s", pod.metadata.name)
            elif pod.status.phase != "Failed":
                continue
            if pod.status.reason == "UnexpectedAdmissionError":
                # This can happen if a non-AdaptDL pod claims the node's
                # resources before this pod could bind to that node.
                LOG.warning("UnexpectedAdmissionError for pod %s: %s",
                            pod.metadata.name, pod.status.message)
            elif str(pod.status.reason).startswith("Outof"):
                # we might be temporarily out of pods on this node
                LOG.warning(f"Pod {pod.metadata.name} is {pod.status.reason} "
                            f"on {pod.spec.node_name}")
            elif preemptible and (pod.metadata.deletion_timestamp is not None
                                  or any143(pod)):
                # This pod was intentionally terminated.
                LOG.warning(f"Pod {pod.metadata.name} terminated")
            else:
                return {"phase": "Failed", "reason": "PodFailure",
                        "message": f"{pod.metadata.name} {pod.status.phase}"}
        return {}

    def _detect_restart(self, pods, allocation):
        for pod in pods:
            replicas = int(pod.metadata.annotations["adaptdl/replicas"])
            rank = int(pod.metadata.annotations["adaptdl/rank"])
            if replicas != len(allocation) or \
                    (pod.metadata.annotations["adaptdl/node"] !=
                     allocation[rank]):
                return True
        return False

    async def _delete_pods(self, pods):
        if not pods:
            return
        results, names = [], []
        for pod in pods:
            if not pod.metadata.deletion_timestamp:
                results.append(self._core_api.delete_namespaced_pod(
                               pod.metadata.name, pod.metadata.namespace))
                names.append(pod.metadata.name)
        if results:
            LOG.info(f"Deleting {names}")
            await asyncio.gather(*results, return_exceptions=True)

    async def _create_pod(self, job_metadata, pod_template,
                          allocation, group, rank):
        node = await self._core_api.read_node(allocation[rank])
        pod = copy.deepcopy(pod_template)
        pod["apiVersion"] = "v1"
        pod["kind"] = "Pod"
        pod.setdefault("metadata", {})
        pod["metadata"]["name"] = self._get_pod_name(job_metadata, group, rank)
        pod["metadata"].setdefault("labels", {})
        pod["metadata"]["ownerReferences"] = \
            templates.owner_reference_template(
                job_metadata["namespace"], job_metadata["name"],
                job_metadata["uid"], kind="AdaptDLJob")
        pod["metadata"]["labels"]["adaptdl"] = "true"
        pod["metadata"]["labels"]["adaptdl/job"] = job_metadata["name"]
        pod["metadata"]["labels"]["petuum.com/nodegroup"] = "all"
        pod["metadata"].setdefault("annotations", {})
        pod["metadata"]["annotations"]["adaptdl/replicas"] = \
            str(len(allocation))
        pod["metadata"]["annotations"]["adaptdl/group"] = str(group)
        pod["metadata"]["annotations"]["adaptdl/rank"] = str(rank)
        pod["metadata"]["annotations"]["adaptdl/node"] = node.metadata.name
        pod["spec"].setdefault("nodeSelector", {})
        pod["spec"]["hostname"] = f"{job_metadata['name']}-{group}-{rank}"
        pod["spec"]["nodeSelector"]["kubernetes.io/hostname"] = \
            node.metadata.labels["kubernetes.io/hostname"]
        pod["spec"]["restartPolicy"] = "Never"
        pod["spec"].setdefault("volumes", [])
        pod["spec"]["volumes"].append({
            "name": "adaptdl-shm",
            "emptyDir": {
                "medium": "Memory",
            },
        })
        pod["spec"] = set_default_resources(pod["spec"])
        for idx, container in enumerate(pod["spec"]["containers"]):
            container.setdefault("volumeMounts", [])
            container["volumeMounts"].append({
                "name": "adaptdl-shm",
                "mountPath": "/dev/shm",
            })
            container.setdefault("env", [])
            container["env"].append({
                "name": "ADAPTDL_JOB_ID",
                "value": "{}/{}".format(job_metadata["namespace"],
                                        job_metadata["name"]),
            })
            container["env"].append({
                "name": "ADAPTDL_MASTER_PORT",
                "value": str(47000 + group),
            })
            container["env"].append({
                "name": "ADAPTDL_NUM_NODES",
                "value": str(len(set(allocation))),
            })
            container["env"].append({
                "name": "ADAPTDL_NUM_RESTARTS",
                "value": str(group),
            })
            container["env"].append({
                "name": "ADAPTDL_NUM_REPLICAS",
                "value": str(len(allocation)),
            })
            container["env"].append({
                "name": "ADAPTDL_REPLICA_RANK",
                "value": str(rank),
            })
            container["env"].append({
                "name": "ADAPTDL_SUPERVISOR_URL",
                "value": config.get_supervisor_url(),
            })
            container["env"].append({
                "name": "ADAPTDL_SCHED_VERSION",
                "value": config.get_adaptdl_version(),
            })
            resources = container.get("resources", {})
            if not resources.get("limits", {}).get("nvidia.com/gpu"):
                # Apparently if a container doesn't ask for any nvidia.com/gpu,
                # then it will be allocated all of them???
                # https://github.com/NVIDIA/k8s-device-plugin/issues/61
                # https://github.com/NVIDIA/k8s-device-plugin
                # AdaptDL overrides this behavior.
                container["env"].append({
                    "name": "NVIDIA_VISIBLE_DEVICES",
                    "value": "none",
                })
        pod = self._patch_pods_and_containers(pod)
        return await self._core_api.create_namespaced_pod(
            job_metadata["namespace"], pod)

    def _patch_pods_and_containers(self, pod):
        pod_patch = config.get_job_patch_pods()
        container_patch = config.get_job_patch_containers()
        if pod_patch:
            pod = jsonpatch.apply_patch(pod, pod_patch)
        if container_patch:
            for idx, container in enumerate(pod["spec"]["containers"]):
                pod["spec"]["containers"][idx] = (
                    jsonpatch.apply_patch(container, container_patch))
        return pod

    def _get_pod_name(self, job_metadata, group, rank):
        job_name = job_metadata["name"]
        job_uid = job_metadata["uid"]
        return f"{job_name}-{job_uid}-{group}-{rank}"



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class JobInfo(object):
    def __init__(self, resources, speedup_fn, creation_timestamp,
                 min_replicas, max_replicas, preemptible=True):
        """
        Args:
            resources (dict): Requested resources (eg. GPUs) of each replica.
            speedup_fn (SpeedupFunction): Speedup function for this job.
            creation_timestamp (datetime): Time when this job was created.
            min_replicas (int): Minimum number of replicas job's guaranteed.
            max_replicas (int): Maximum number of replicas. Maximum should be
                                greater or equal to Minimum
            preemptible (bool): Is the job preemptible?
        """
        assert max_replicas > 0
        assert max_replicas >= min_replicas
        self.resources = resources
        self.speedup_fn = speedup_fn
        self.creation_timestamp = creation_timestamp
        self.max_replicas = max_replicas
        self.min_replicas = min_replicas
        self.preemptible = preemptible


class NodeInfo(object):
    def __init__(self, resources, preemptible):
        """
        Args:
            resources (dict): Available resources (eg. GPUs) on this node.
            preemptible (bool): Whether this node is pre-emptible.
        """
        self.resources = resources
        self.preemptible = preemptible



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import kubernetes_asyncio as kubernetes
from aiohttp import web
import logging
from adaptdl.sched_hints import SCHED_HINTS
from adaptdl_sched.config import get_supervisor_port


LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class Supervisor:
    """
    Supervisor provides a simple REST interface for several functionalities.
    Currently, it has two endpoints:
    1.  /hints for jobs to send scheduling hints.
    2.  /discover for finding the pod IPs of a job.
    """

    def __init__(self, port, host='0.0.0.0'):
        self._host = host
        self._port = port
        self._core_api = kubernetes.client.CoreV1Api()
        self._objs_api = kubernetes.client.CustomObjectsApi()

    async def _handle_healthz(self, request):
        # Health check.
        return web.Response()

    async def _handle_discover(self, request):
        # Long-polling endpoint used for discovering pod IPs for a given job.
        namespace = request.match_info["namespace"]
        name = request.match_info["name"]
        group = request.match_info["group"]
        timeout = int(request.query.get("timeout", "30"))
        pod_ip_list = None
        async with kubernetes.watch.Watch() as w:
            stream = w.stream(self._core_api.list_namespaced_pod, namespace,
                              label_selector="adaptdl/job={}".format(name),
                              field_selector="status.podIP!=",
                              timeout_seconds=timeout)
            async for event in stream:
                pod = event["object"]
                replicas = int(pod.metadata.annotations["adaptdl/replicas"])
                rank = int(pod.metadata.annotations["adaptdl/rank"])
                if pod.metadata.annotations["adaptdl/group"] == group:
                    if pod_ip_list is None:
                        pod_ip_list = [None] * replicas
                    pod_ip_list[rank] = pod.status.pod_ip
                    if all(pod_ip is not None for pod_ip in pod_ip_list):
                        return web.json_response(pod_ip_list)
        return web.json_response(status=408)  # Timeout.

    async def _handle_report(self, request):
        namespace = request.match_info['namespace']
        name = request.match_info['name']
        hints = await request.json()
        # Drop all unrecognized fields. TODO: validate each client-sent field.
        hints = {k: hints[k] for k in SCHED_HINTS if k in hints}
        # Patch only the train field to avoid conflicts with controller.
        patch = {"status": {"train": hints}}
        LOG.info("Patch AdaptDLJob %s/%s: %s", namespace, name, patch)
        await self._objs_api.patch_namespaced_custom_object_status(
            "adaptdl.petuum.com", "v1", namespace, "adaptdljobs", name, patch)
        return web.Response()

    def run(self):
        self.app = web.Application()
        self.app.add_routes([
            web.get('/healthz', self._handle_healthz),
            web.get('/discover/{namespace}/{name}/{group}',
                    self._handle_discover),
            web.put('/hints/{namespace}/{name}', self._handle_report),
        ])
        LOG.info("%s %s", self._host, self._port)
        web.run_app(self.app, host=self._host, port=self._port)


if __name__ == "__main__":
    logging.basicConfig()
    kubernetes.config.load_incluster_config()

    supervisor = Supervisor(get_supervisor_port())
    supervisor.run()



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import adaptdl_sched.resources as resources
import copy
import os


def test_discretize_resource_cpu1():
    name = "cpu"
    value = "10k"
    assert(resources._discretize_resource(name, value) == 1e7)


def test_discretize_resource_cpu2():
    name = "cpu"
    value = "100m"
    assert(resources._discretize_resource(name, value) == 100)


def test_discretize_resource_cpu3():
    name = "cpu"
    value = 1000
    assert(resources._discretize_resource(name, value) == 1e6)


def test_discretize_resource_cpu4():
    name = "cpu"
    value = "1M"
    assert(resources._discretize_resource(name, value) == 1e9)


def test_discretize_resource_cpu5():
    name = "cpu"
    value = "1.37m"
    assert(resources._discretize_resource(name, value) == 2)


def test_discretize_resource_mem1():
    name = "memory"
    value = "1T"
    assert(resources._discretize_resource(name, value) == 1e12)


def test_discretize_resource_mem2():
    name = "memory"
    value = "1Gi"
    assert(resources._discretize_resource(name, value) == 1024 ** 3)


def test_discretize_resource_mem3():
    name = "memory"
    value = "10e3"
    assert(resources._discretize_resource(name, value) == 10000)


pod_spec = {
        'containers': [{
            'image': 'foo',
            'name': 'foo',
            'command': ['foo'],
            'resources': {}
        }],
        'apiVersion': 'v1',
        'kind': 'Pod',
        'metadata': {
            'name': 'foo'
        }
    }


def test_set_default_resources1():
    os.environ['ADAPTDL_JOB_DEFAULT_RESOURCES'] = \
        '{"limits": {"memory": 100}, "requests": {"memory": 200}}'
    pod_spec_copy = copy.deepcopy(pod_spec)
    pod_spec_copy = resources.set_default_resources(pod_spec_copy)
    res = pod_spec_copy['containers'][0]['resources']
    assert(res == {"limits": {"memory": 100}, "requests": {"memory": 200}})


def test_set_default_resources2():
    os.environ['ADAPTDL_JOB_DEFAULT_RESOURCES'] = \
        '{"limits": {"memory": 100}, "requests": {"memory": 200}}'
    pod_spec_copy = copy.deepcopy(pod_spec)
    pod_spec_copy['containers'][0]['resources'] = {"limits": {"memory": 300}}
    pod_spec_copy = resources.set_default_resources(pod_spec_copy)
    res = pod_spec_copy['containers'][0]['resources']
    assert(res == {"limits": {"memory": 300}, "requests": {"memory": 200}})


def test_set_default_resources3():
    os.environ['ADAPTDL_JOB_DEFAULT_RESOURCES'] = \
        '{"requests": {"memory": 200}}'
    pod_spec_copy = copy.deepcopy(pod_spec)
    pod_spec_copy['containers'][0]['resources'] = {"limits": {"memory": 300}}
    pod_spec_copy = resources.set_default_resources(pod_spec_copy)
    res = pod_spec_copy['containers'][0]['resources']
    assert(res == {"limits": {"memory": 300}, "requests": {"memory": 200}})


def test_set_default_resources4():
    os.environ['ADAPTDL_JOB_DEFAULT_RESOURCES'] = \
        '{"limits": {"memory": 200}}'
    pod_spec_copy = copy.deepcopy(pod_spec)
    pod_spec_copy['containers'][0]['resources'] = {"limits": {"memory": 300}}
    pod_spec_copy = resources.set_default_resources(pod_spec_copy)
    res = pod_spec_copy['containers'][0]['resources']
    assert(res == {"limits": {"memory": 300}})


pod_spec_2 = {
    'containers': [
        {
            'image': 'foo',
            'name': 'foo',
            'command': ['foo'],
            'resources': {
                'requests': {
                    'memory': 100,
                    'cpu': 1.5
                },
                'limits': {
                    'memory': 80,
                    'cpu': 1,
                    'cows': 2
                }
            }
        }, {
            'image': 'foo2',
            'name': 'foo2',
            'command': ['foo2'],
            'resources': {
                'requests': {
                    'memory': 120,
                    'cpu': 2.5
                },
                'limits': {
                    'memory': 100,
                    'cpu': 2,
                    'cows': 5
                }
            }
        }, {
            'image': 'foo3',
            'name': 'foo3',
            'command': ['foo3'],
            'resources': {
                'requests': {
                    'memory': 250,
                    'cpu': 10
                },
                'limits': {
                    'memory': 250,
                    'cpu': 10
                }
            }
        }
    ],
    'apiVersion': 'v1',
    'kind': 'Pod',
    'metadata': {
        'name': 'foo'
    }
}


def test_get_pod_requests():
    pod_spec_copy = copy.deepcopy(pod_spec_2)
    requests = resources.get_pod_requests(pod_spec_copy)
    assert(requests == {
        "pods": 1,
        "cpu": resources._discretize_resource("cpu", 14),
        "memory": resources._discretize_resource("memory", 470),
        "cows": resources._discretize_resource("cows", 7)})



import json
import kubernetes_asyncio as kubernetes
import uuid

from http import HTTPStatus
from unittest.mock import AsyncMock

from adaptdl_sched.validator import Validator


def _assert_response(response_json, request_json):
    assert response_json["apiVersion"] == "admission.k8s.io/v1"
    assert response_json["kind"] == "AdmissionReview"
    assert response_json["response"]["uid"] == request_json["request"]["uid"]


async def test_healthz(aiohttp_client, loop):
    app = Validator().get_app()
    client = await aiohttp_client(app)
    response = await client.get("/healthz")
    assert response.status == HTTPStatus.OK


async def test_create_invalid_template(aiohttp_client, loop):
    validator = Validator()
    # Set up mocks.
    exc = kubernetes.client.rest.ApiException(
        status=HTTPStatus.UNPROCESSABLE_ENTITY, reason="reason")
    exc.body = json.dumps({"message": str(uuid.uuid4())})
    mock = AsyncMock(side_effect=exc)
    validator._core_api.create_namespaced_pod_template = mock
    # Send request.
    app = validator.get_app()
    client = await aiohttp_client(app)
    template = {"key": str(uuid.uuid4())}
    request_json = {
        "request": {
            "uid": str(uuid.uuid4()),
            "operation": "CREATE",
            "namespace": str(uuid.uuid4()),
            "object": {"spec": {"template": template}},
        }
    }
    response = await client.post("/validate", json=request_json)
    # Check template dry run.
    assert mock.call_args.args[0] == request_json["request"]["namespace"]
    assert mock.call_args.args[1]["template"] == template
    assert mock.call_args.kwargs["dry_run"] == "All"
    # Check HTTP response.
    assert response.status == HTTPStatus.OK
    response_json = await response.json()
    _assert_response(response_json, request_json)
    # Check operation was disallowed.
    assert not response_json["response"]["allowed"]
    status = response_json["response"]["status"]
    assert status["code"] == HTTPStatus.UNPROCESSABLE_ENTITY
    assert status["reason"] == "Invalid"
    assert status["message"] == json.loads(exc.body)["message"]


async def test_create_invalid_replicas(aiohttp_client, loop):
    validator = Validator()
    validator._core_api.create_namespaced_pod_template = AsyncMock()
    # Send request.
    app = validator.get_app()
    client = await aiohttp_client(app)
    request_json = {
        "request": {
            "uid": str(uuid.uuid4()),
            "operation": "CREATE",
            "namespace": str(uuid.uuid4()),
            "object": {"spec": {"minReplicas": 4, "maxReplicas": 2,
                                "template": {}}},
        }
    }
    response = await client.post("/validate", json=request_json)
    # Check HTTP response.
    assert response.status == HTTPStatus.OK
    response_json = await response.json()
    _assert_response(response_json, request_json)
    # Check operation was disallowed.
    assert not response_json["response"]["allowed"]
    status = response_json["response"]["status"]
    assert status["code"] == HTTPStatus.UNPROCESSABLE_ENTITY
    assert status["reason"] == "Invalid"


async def test_update_spec(aiohttp_client, loop):
    validator = Validator()
    # Send request.
    app = validator.get_app()
    client = await aiohttp_client(app)
    request_json = {
        "request": {
            "uid": str(uuid.uuid4()),
            "operation": "UPDATE",
            "namespace": str(uuid.uuid4()),
            "object": {"spec": {"key": "value"}},
            "oldObject": {"spec": {"key": "oldValue"}},
        }
    }
    response = await client.post("/validate", json=request_json)
    # Check HTTP response.
    assert response.status == HTTPStatus.OK
    response_json = await response.json()
    _assert_response(response_json, request_json)
    # Check operation was disallowed.
    assert not response_json["response"]["allowed"]
    status = response_json["response"]["status"]
    assert status["code"] == HTTPStatus.UNPROCESSABLE_ENTITY
    assert status["reason"] == "Forbidden"



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from unittest.mock import Mock
from adaptdl_sched.policy.speedup import SpeedupFunction


def mock_optimize(num_nodes, num_replicas, *args, **kwargs):
    return 32 * np.sqrt(num_replicas), 32, 0


def test_speedup():
    goodput_fn = Mock()
    goodput_fn.optimize = Mock(side_effect=mock_optimize)
    speedup_fn = SpeedupFunction(goodput_fn)
    num_replicas = np.arange(1, 100)
    expected_ret = np.sqrt(num_replicas)
    assert np.allclose(speedup_fn(1, num_replicas), expected_ret)


def test_memoize():
    goodput_fn = Mock()
    goodput_fn.optimize = Mock(side_effect=mock_optimize)
    speedup_fn = SpeedupFunction(goodput_fn)
    goodput_fn.optimize.reset_mock()
    # Call with never-seen inputs.
    num_nodes = num_replicas = np.arange(1, 10)
    result = speedup_fn(num_nodes, num_replicas)
    assert np.allclose(result, np.sqrt(num_replicas))
    assert goodput_fn.optimize.call_count == 1
    assert np.all(goodput_fn.optimize.call_args[0][0] == num_nodes)
    assert np.all(goodput_fn.optimize.call_args[0][1] == num_replicas)
    # Call with some-seen inputs.
    num_nodes = num_replicas = np.arange(1, 20)
    result = speedup_fn(num_nodes, num_replicas)
    assert np.allclose(result, np.sqrt(num_replicas))
    assert goodput_fn.optimize.call_count == 2
    assert np.all(goodput_fn.optimize.call_args[0][0] == np.arange(10, 20))
    assert np.all(goodput_fn.optimize.call_args[0][1] == np.arange(10, 20))
    # Call with all-seen inputs.
    num_nodes_3 = num_replicas_3 = np.arange(5, 15)
    expect_3 = np.sqrt(num_replicas_3)
    result_3 = speedup_fn(num_nodes_3, num_replicas_3)
    assert np.allclose(result_3, expect_3)
    assert goodput_fn.optimize.call_count == 2  # Shouldn't have increased.



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import time

from collections import Counter
from datetime import datetime, timedelta
from adaptdl.goodput import GoodputFunction, PerfParams, GradParams
from adaptdl_sched.policy.pollux import PolluxPolicy
from adaptdl_sched.policy.speedup import SpeedupFunction
from adaptdl_sched.policy.utils import JobInfo, NodeInfo


@pytest.mark.parametrize("num_nodes", [1, 2, 4, 8, 16])
def test_optimize(num_nodes, total_devices=16):
    assert total_devices % num_nodes == 0
    num_devices = total_devices // num_nodes
    print("{}x{} nodes:".format(num_nodes, num_devices))
    # Make up a realistic speedup function.
    perf_params = PerfParams(0.121, 0.00568, 0.0236, 0.00634,
                             0.0118, 0.00317, 1.14)
    grad_params = GradParams(sqr=0.00136, var=0.000502)
    goodput_fn = GoodputFunction(perf_params, grad_params, 128)
    speedup_fn = SpeedupFunction(goodput_fn, max_batch_size=1280,
                                 atomic_bsz_range=(64, 256))
    now = datetime.now()
    jobs = {}
    # Add a few jobs.
    job_resources = {"nvidia.com/gpu": 1, "pods": 1}
    for i in range(16):
        creation_timestamp = now + timedelta(minutes=len(jobs)),
        max_replicas = 8
        min_replicas = 0
        key = len(jobs)
        jobs[key] = JobInfo(job_resources, speedup_fn, creation_timestamp,
                            min_replicas, max_replicas)
    # Add a few nodes.
    node_resources = {"nvidia.com/gpu": num_devices, "pods": 32}
    nodes = {i: NodeInfo(node_resources, preemptible=False)
             for i in range(num_nodes)}
    # Add a node template.
    node_template = NodeInfo(node_resources, preemptible=True)
    policy = PolluxPolicy()
    prev_allocs = {}
    for i in range(3):
        start = time.time()
        allocations, desired_nodes = \
            policy.optimize(jobs, nodes, prev_allocs, node_template)
        duration = time.time() - start
        print("optimize {}x ({}s sec):".format(i + 1, duration))
        node_count = Counter()
        for job_key, placement in allocations.items():
            assert len(placement) <= jobs[job_key].max_replicas
            for node_key in placement:
                node_count[node_key] += 1
        for node_key, count in node_count.items():
            assert count <= nodes[node_key].resources["nvidia.com/gpu"]
            assert count <= nodes[node_key].resources["pods"]


def test_allocate_job():
    nodes = {
        "0": NodeInfo({"gpu": 1, "cpu": 500, "pods": 32}, preemptible=False),
        "1": NodeInfo({"gpu": 2, "cpu": 2000, "pods": 32}, preemptible=False),
        "2": NodeInfo({"gpu": 2, "cpu": 3000, "pods": 32}, preemptible=True),
    }
    perf_params = PerfParams(0.121, 0.00568, 0.0236, 0.00634,
                             0.0118, 0.00317, 1.14)
    grad_params = GradParams(sqr=0.00136, var=0.000502)
    goodput_fn = GoodputFunction(perf_params, grad_params, 128)
    speedup_fn = SpeedupFunction(goodput_fn, max_batch_size=1280,
                                 atomic_bsz_range=(64, 256))
    now = datetime.now()
    min_replicas = 0
    job_1 = JobInfo({"gpu": 1, "cpu": 500, "pods": 1}, speedup_fn,
                    now + timedelta(minutes=0), min_replicas, max_replicas=1)
    job_2 = JobInfo({"gpu": 1, "cpu": 1000, "pods": 1}, speedup_fn,
                    now + timedelta(minutes=1), min_replicas, max_replicas=1)
    job_3 = JobInfo({"gpu": 1, "cpu": 1000, "pods": 1}, speedup_fn,
                    now + timedelta(minutes=1), 2, max_replicas=2)
    job_4 = JobInfo({"gpu": 1, "cpu": 2000, "pods": 1}, speedup_fn,
                    now + timedelta(minutes=1), 2, max_replicas=2)
    policy = PolluxPolicy()

    assert(policy.allocate_job(job_1, nodes) == ["0"])
    assert(policy.allocate_job(job_2, nodes) == ["1"])
    assert(policy.allocate_job(job_3, nodes) == ["1", "1"])
    assert(policy.allocate_job(job_4, nodes) == [])


def test_unusable_node():
    # Test where one of the nodes can't be used due to one resource type.
    nodes = {
        0: NodeInfo({"gpu": 1, "cpu": 500, "pods": 32}, preemptible=False),
        1: NodeInfo({"gpu": 1, "cpu": 8000, "pods": 32}, preemptible=False),
        2: NodeInfo({"gpu": 1, "cpu": 8000, "pods": 32}, preemptible=False),
    }
    template = NodeInfo({"gpu": 1, "cpu": 8000, "pods": 32}, preemptible=True)
    perf_params = PerfParams(0.121, 0.00568, 0.0236, 0.00634,
                             0.0118, 0.00317, 1.14)
    grad_params = GradParams(sqr=0.00136, var=0.000502)
    goodput_fn = GoodputFunction(perf_params, grad_params, 128)
    speedup_fn = SpeedupFunction(goodput_fn, max_batch_size=1280,
                                 atomic_bsz_range=(64, 256))
    now = datetime.now()
    min_replicas = 0
    jobs = {
        0: JobInfo({"gpu": 1, "cpu": 1000, "pods": 1}, speedup_fn,
                   now + timedelta(minutes=0), min_replicas, max_replicas=1),
        1: JobInfo({"gpu": 1, "cpu": 1000, "pods": 1}, speedup_fn,
                   now + timedelta(minutes=1), min_replicas, max_replicas=1),
        2: JobInfo({"gpu": 1, "cpu": 1000, "pods": 1}, speedup_fn,
                   now + timedelta(minutes=2), min_replicas, max_replicas=1),
    }
    policy = PolluxPolicy()
    allocations, desired_nodes = policy.optimize(jobs, nodes, {}, template)
    # Check that more nodes are asked for.
    assert desired_nodes > 3
    # Check no job was allocated more than 1 replica.
    assert max(len(alloc) for alloc in allocations.values()) == 1
    # Check two jobs were allocated.
    assert sum(len(alloc) for alloc in allocations.values()) == 2



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import kubernetes_asyncio as kubernetes
import logging
import prometheus_client

from adaptdl_sched.controller import AdaptDLController

logging.basicConfig()
kubernetes.config.load_incluster_config()
prometheus_client.start_http_server(9091)

controller = AdaptDLController()

loop = asyncio.get_event_loop()
loop.run_until_complete(
    controller.run(),
)
loop.close()



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import kubernetes_asyncio as kubernetes
from kubernetes_asyncio.client.rest import ApiException
import logging
import adaptdl_sched.config as config
from collections import namedtuple
import adaptdl_sched.k8s_templates as templates

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class ClusterExpander(object):
    """ ClusterExpander tries to keep expected node count available to the
    allocator. It does that by spawning equal number of placeholder pods (one
    of each node). The pods have anti-affinity which prevents them from getting
    scheduled on the same node. This pushes the cluster autoscaler to provision
    one node for each Pending placeholder. The job of ClusterExpander then is
    really to maintain the requested count of placeholders. This also holds
    when the jobs are scaling down or finshing and we no longer need specific
    nodes"""

    def __init__(self):
        self._v1 = kubernetes.client.CoreV1Api()
        self._apps_api = kubernetes.client.AppsV1Api()
        self._allocations = set()
        self._active_nodes = set()

    def _gen_placeholder_pod(self):
        return {
                'apiVersion': 'v1',
                'kind': 'Pod',
                'metadata': {
                    'generateName': 'adaptdl-placeholder-',
                    'labels': {
                        config.ADAPTDL_PH_LABEL: 'true',
                        'petuum.com/nodegroup': 'adaptdl'
                        },
                    'ownerReferences': self._owner_reference
                    },
                'spec': {
                    'affinity': {
                        'podAntiAffinity': {
                            'requiredDuringSchedulingIgnoredDuringExecution': [
                                {
                                    'labelSelector': {
                                        'matchExpressions': [
                                            {
                                                'key': config.ADAPTDL_PH_LABEL,
                                                'operator': 'In',
                                                'values': ['true']
                                            }]
                                        },
                                    'topologyKey': 'kubernetes.io/hostname'
                                    }
                                ]
                            }
                    },
                    'containers': [{
                        'image': 'busybox',
                        'name': 'placeholder',
                        'command': ["/bin/sh", "-ec",
                                    "while :; do echo '.'; sleep 5 ; done"],
                        "resources": {
                            "requests": {
                                "memory": "5Mi",
                                "cpu": "1m"
                            }
                        }
                    }],
                    'restartPolicy': 'Never',
                    }
                }

    async def _reconcile(self, expected):
        Pod = namedtuple('Pod', ['name', 'node', 'phase'])

        def key_fn(x):
            if x.node in self._allocations and x.phase == 'Running':
                return 2
            elif x.node not in self._allocations and x.phase == 'Running':
                return 1
            else:
                return 0

        try:
            ret = await self._v1.list_namespaced_pod(config.get_namespace(),
                                                     label_selector=f"{config.ADAPTDL_PH_LABEL}=true")  # noqa: E501

            pods = [Pod(pod.metadata.name, pod.spec.node_name,
                        pod.status.phase) for pod in ret.items]

            # sort pods based on 1. in allocations 2. Running 3. Pending
            pods = sorted(pods, key=key_fn)

            running = sum(1 if pod.phase == 'Running' else 0
                          for pod in pods)
            spawned = sum(1 if pod.phase in ('Running', 'Pending') else 0
                          for pod in pods)

            succeeded = sum(1 if pod.phase == 'Succeeded' else 0
                            for pod in pods)
            assert succeeded == 0

            LOG.info(f"Received {expected} expected nodes. There are "
                     f"{running} Running, {spawned - running} Pending "
                     f"placeholder pods.")

            if expected == spawned:
                return
            elif expected > spawned:  # spawn the difference
                r = []
                for _ in range(expected - spawned):
                    r.append(self._v1.create_namespaced_pod(
                        body=self._gen_placeholder_pod(),
                        namespace=config.get_namespace()))
                ret = await asyncio.gather(*r)
                for pod in ret:
                    LOG.info(f"Spawned {pod.metadata.name}")
            else:   # expected < spawned, delete from the left
                r = []
                for name in [x.name for x in pods][:spawned - expected]:
                    assert name not in self._allocations
                    r.append(self._v1.delete_namespaced_pod(
                        name=name, namespace=config.get_namespace()))
                await asyncio.gather(*r)
                for name in [x.name for x in pods][:spawned - expected]:
                    LOG.info(f"Deleted {name}")
        except ApiException as e:
            if e.status == 401:  # Unauthorized
                raise  # Fatal
            else:
                LOG.warning(f"{e}")

    async def run(self):
        adaptdl_sched = \
            await self._apps_api.read_namespaced_deployment(
                namespace=config.get_namespace(),
                name=config.get_adaptdl_deployment())
        self._owner_reference = templates.owner_reference_template(
            config.get_namespace(),
            adaptdl_sched.metadata.name,
            adaptdl_sched.metadata.uid,
            "Deployment",
            "apps/v1")
        while True:
            await self._reconcile(len(self._active_nodes))
            await asyncio.sleep(30)

    def fit(self, active_nodes):
        """ active_nodes contain allocations + virtual nodes, our job at the
        expander is to 1. maintain the allocations, and 2. provision nodes in
        active nodes and not in allocations."""
        self._active_nodes = set(active_nodes)
        self._allocations = set()
        for node in self._active_nodes:
            if not node.startswith("~"):  # real nodes only
                self._allocations.add(node)


if __name__ == '__main__':
    # unit test to check basic sanity
    loop = asyncio.get_event_loop()
    loop.run_until_complete(kubernetes.config.load_kube_config())

    expander = ClusterExpander()

    async def run():
        expander.fit(['n0', '~n1'])
        await expander.run()

    loop.run_until_complete(run())
    loop.close()



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


class SpeedupFunction(object):

    def __init__(self, goodput_fn, max_batch_size=None, atomic_bsz_range=None,
                 accumulation=False, mem_size=32):
        self._goodput_fn = goodput_fn
        self._max_batch_size = max_batch_size
        self._atomic_bsz_range = atomic_bsz_range
        self._accumulation = accumulation
        self._mem_size = mem_size
        self._base_goodput, _, _ = goodput_fn.optimize(
            num_nodes=1, num_replicas=1, max_batch_size=max_batch_size,
            atomic_bsz_range=atomic_bsz_range, accumulation=accumulation)
        # Memoization for fast repeated queries.
        self._mem_speedup = -np.ones((mem_size, mem_size))
        self._mem_speedup[0, 0] = 0.0

    def __call__(self, num_nodes, num_replicas):
        assert np.all(np.less_equal(0, num_nodes))
        assert np.all(np.less_equal(num_nodes, num_replicas))
        assert np.all((num_nodes > 0) == (num_replicas > 0))
        # Remember what the output shape/format should be and flatten inputs.
        output_scalar = np.isscalar(num_nodes) and np.isscalar(num_replicas)
        output_shape = np.broadcast(num_nodes, num_replicas).shape
        num_nodes = np.broadcast_to(num_nodes, output_shape).flatten()
        num_replicas = np.broadcast_to(num_replicas, output_shape).flatten()
        # Return values which will be filled out.
        speedup = -np.ones(output_shape).flatten()
        # Fill in any previously memoized results first.
        indices = num_replicas < self._mem_size
        mem_idx = (num_nodes[indices], num_replicas[indices])
        speedup[indices] = self._mem_speedup[mem_idx]
        # Find the missing indices which still need to be computed.
        missing = speedup < 0
        if np.count_nonzero(missing) > 0:
            num_nodes, num_replicas = num_nodes[missing], num_replicas[missing]
            # Find unique inputs to reduce compuation.
            (num_nodes, num_replicas), inverse = np.unique(
                    np.stack([num_nodes, num_replicas]),
                    axis=1, return_inverse=True)
            goodput, _, _ = self._goodput_fn.optimize(
                num_nodes, num_replicas,
                max_batch_size=self._max_batch_size,
                atomic_bsz_range=self._atomic_bsz_range,
                accumulation=self._accumulation)
            # Memoize results.
            indices = num_replicas < self._mem_size
            mem_idx = (num_nodes[indices], num_replicas[indices])
            self._mem_speedup[mem_idx] = goodput[indices] / self._base_goodput
            # Fill in computed results.
            speedup[missing] = goodput[inverse] / self._base_goodput
        assert np.all(np.less_equal(0, speedup))
        speedup = speedup.reshape(output_shape)
        return speedup.item() if output_scalar else speedup



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import kubernetes_asyncio as kubernetes


async def patch_job_status(obj_api, namespace, name, patch):
    try:
        return await obj_api.patch_namespaced_custom_object_status(
            "adaptdl.petuum.com", "v1", namespace, "adaptdljobs", name, patch)
    except kubernetes.client.rest.ApiException as exc:
        if exc.status == 404:
            return None
        raise



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import dateutil.parser
import kubernetes_asyncio as kubernetes
import logging
import time

from adaptdl.goodput import GoodputFunction, PerfParams, GradParams
from adaptdl.sched_hints import PERF_PARAMS
from adaptdl_sched.policy.pollux import PolluxPolicy
from adaptdl_sched.policy.speedup import SpeedupFunction
from adaptdl_sched.policy.utils import JobInfo, NodeInfo
from adaptdl_sched.resources import (get_node_unrequested, get_pod_requests,
                                     set_default_resources)
from adaptdl_sched.utils import patch_job_status
from adaptdl_sched.cluster_expander import ClusterExpander
from adaptdl_sched.config import allowed_taints

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class AdaptDLAllocator(object):
    def __init__(self, expander):
        self._core_api = kubernetes.client.CoreV1Api()
        self._objs_api = kubernetes.client.CustomObjectsApi()
        self._custom_resource = ("adaptdl.petuum.com", "v1",
                                 "", "adaptdljobs")
        self._cluster_expander = expander
        self._policy = PolluxPolicy()
        # lock for the two corountines in run()
        self._lock = asyncio.Lock()

    async def run(self):
        # two functionality: (1) watch for new job and start if possible.
        # (2) periodically optimize existing jobs
        await asyncio.gather(
            self._allocate_one_loop(),
            self._optimize_all_loop()
        )

    async def _allocate_one_loop(self):
        async with kubernetes.watch.Watch() as watch:
            while True:
                async for event in watch.stream(
                        self._objs_api.list_namespaced_custom_object,
                        *self._custom_resource, timeout_seconds=60):
                    # We only consider newly-added preemptible jobs
                    # because this allocation may not be final.
                    if (event["type"] == "ADDED" and
                            event["object"]["spec"].get("preemptible", True)):
                        async with self._lock:
                            await self._allocate_one(event)

    async def _allocate_one(self, event):
        # re-read the job , compared with the previous readjob
        job = event["object"]
        namespace = job["metadata"]["namespace"]
        name = job["metadata"]["name"]

        try:
            job = await self._objs_api.get_namespaced_custom_object(
                "adaptdl.petuum.com", "v1", namespace, "adaptdljobs", name
            )
        except kubernetes.client.rest.ApiException as exc:
            if exc.status == 404:
                return
            raise  # unexpected

        # some other coroutine has handled this
        if job.get("status", {}).get("allocation") is not None or\
                job.get("status", {}).get("group") is not None:
            return

        namespace = job["metadata"]["namespace"]
        name = job["metadata"]["name"]
        # if this is a restarted job, skip i
        LOG.info("detected an added job %s/%s.",
                 namespace, name)
        # parse the job infomation
        job_info = self._get_job_info(job)

        # find available nodes.
        node_infos, _ = await self._find_nodes()

        # get the node to allocate
        new_allocation = self._policy.allocate_job(
                            job_info, node_infos)
        patch = {"status": {"allocation": new_allocation}}
        LOG.info("Patch AdaptdlJob %s/%s: %s ",
                 namespace, name, patch)
        await patch_job_status(self._objs_api, namespace, name, patch)

    async def _optimize_all_loop(self):
        while True:
            # try to gain lock
            async with self._lock:
                await self._optimize_all()

            LOG.info("Sleep for 60 seconds")
            await asyncio.sleep(60)

    async def _optimize_all(self):
        LOG.info("Running allocator loop")
        nodes, node_template = await self._find_nodes(
                                   pod_label_selector="!adaptdl/job"
                               )
        LOG.info("Node resources: %s",
                 {k: v.resources for k, v in nodes.items()})
        jobs, prev_allocations = \
            await self._find_jobs_and_allocations()
        LOG.info("Job resources: %s",
                 {k: v.resources for k, v in jobs.items()})
        start = time.time()
        allocations = self._allocate(jobs, nodes, prev_allocations,
                                     node_template)
        duration = time.time() - start
        LOG.info("Allocations (in %.3f sec): %s", duration,
                 allocations)
        await self._update_allocations(allocations)

    async def _update_allocations(self, allocations):
        job_list = await self._objs_api.list_namespaced_custom_object(
            "adaptdl.petuum.com", "v1", "", "adaptdljobs")
        for job in job_list["items"]:
            namespace = job["metadata"]["namespace"]
            name = job["metadata"]["name"]
            job_allocation = job.get("status", {}).get("allocation", [])
            new_allocation = list(allocations.get((namespace, name), []))
            if list(job_allocation) != new_allocation:
                patch = {"status": {"allocation": new_allocation}}
                LOG.info("Patch AdaptDLJob %s/%s: %s", namespace, name, patch)
                await patch_job_status(self._objs_api, namespace, name, patch)

    async def _find_nodes(self, pod_label_selector=""):
        # NOTE: Default empty string "" label_selector qualifies all pods
        node_infos = {}
        node_list = await self._core_api.list_node()
        # Find all pods qualified by the pod_label_selector which are taking up
        # resources and subtract those resources from the available pool.
        # Apparently there's not a more efficient way to get currently
        # available resources in k8s?. We also check if we have reached the pod
        # limit on the node. This number denotes (allocatable pods -
        # Non-terminated pods) on that node.
        pod_list = await self._core_api.list_pod_for_all_namespaces(
                                label_selector=pod_label_selector)
        for node in node_list.items:
            if allowed_taints(node.spec.taints):
                resources = get_node_unrequested(node, pod_list.items)
                if not resources.get("pods"):
                    LOG.warning(f"node {node.metadata.name} "
                                "has no free pods available.")
                node_infos[node.metadata.name] = NodeInfo(resources, False)
        # For cluster autoscaling: to determine if additional nodes would be
        # helpful, add a few "virtual" nodes which only become available in
        # "eta" seconds. Currently, we only consider as many virtual nodes as
        # there are real nodes. We infer each resource to be the maximum amount
        # observed in any real node.
        max_resources = {}
        for node_name in node_infos:
            for key, val in node_infos[node_name].resources.items():
                if key not in max_resources or val > max_resources[key]:
                    max_resources[key] = val
        node_template = NodeInfo(max_resources, True)
        return node_infos, node_template

    def _get_job_info(self, job):
        job["spec"]["template"]["spec"] = \
            set_default_resources(job["spec"]["template"]["spec"])
        resources = get_pod_requests(job["spec"]["template"]["spec"])
        hints = job.get("status", {}).get("train", {})
        max_replicas = max(2 * hints.get("maxProfiledReplicas", 0), 1)
        if job["spec"].get("maxReplicas"):
            max_replicas = min(max_replicas, job["spec"]["maxReplicas"])
        min_replicas = job["spec"].get("minReplicas", 0)
        # max_replicas should be greater or equal to min_replicas
        max_replicas = max(max_replicas, min_replicas)
        preemptible = job["spec"].get("preemptible", True)
        if {"perfParams", "initBatchSize"} <= hints.keys() and preemptible:
            max_batch_size = (hints.get("maxBatchSize")
                              or hints["initBatchSize"])
            if hints.get("localBszBounds"):
                min_local_bsz = hints["localBszBounds"][0] or 1
                # Make sure max_batch_size / replicas >= min_local_bsz
                if max_batch_size < min_local_bsz * max_replicas:
                    max_replicas = int(max_batch_size / min_local_bsz)
            perf_params = PerfParams(*[hints["perfParams"][k]
                                       for k in PERF_PARAMS.keys()])
            if "gradParams" in hints:
                grad_params = GradParams(hints["gradParams"]["norm"],
                                         hints["gradParams"]["var"])
            else:
                grad_params = GradParams(0.0, 1.0)
            goodput_fn = GoodputFunction(perf_params, grad_params,
                                         hints["initBatchSize"])
            speedup_fn = SpeedupFunction(
                goodput_fn,
                hints.get("maxBatchSize"),
                hints.get("localBszBounds"),
                hints.get("gradientAccumulation", False))
        else:
            speedup_fn = lambda n, r: r  # noqa: E731
        creation_ts = dateutil.parser.isoparse(
                job["metadata"]["creationTimestamp"])
        return JobInfo(
                resources, speedup_fn, creation_ts, min_replicas,
                max_replicas, preemptible)

    async def _find_jobs_and_allocations(self):
        job_list = await self._objs_api.list_namespaced_custom_object(
            "adaptdl.petuum.com", "v1", "", "adaptdljobs")
        job_infos = {}
        allocations = {}

        for job in job_list["items"]:
            if job.get("status", {}).get("phase") \
                    not in ["Pending", "Running", "Starting", "Stopping"]:
                continue

            namespace = job["metadata"]["namespace"]
            name = job["metadata"]["name"]

            if "allocation" in job.get("status", {}):
                allocations[namespace, name] = \
                    list(job["status"]["allocation"])

            job_info = self._get_job_info(job)
            job_infos[(namespace, name)] = job_info

        return job_infos, allocations

    def _allocate(self, jobs, nodes, prev_allocations, node_template):
        for job_key in list(jobs):
            job_resources = jobs[job_key].resources
            for node in nodes.values():
                if all(val <= node.resources.get(key, 0)
                       for key, val in job_resources.items()):
                    # Found a node which can fit a replica of this job.
                    break
            else:
                # No node can fit a replica of this job.
                # TODO: propagate this to the controller so the job is Failed.
                LOG.warning("Job %s cannot be scheduled!", job_key)
                jobs.pop(job_key)
        allocations = {}
        if not jobs:
            # There are no jobs, let the expander shrink the cluster.
            self._cluster_expander.fit([])
        elif jobs and nodes:
            allocations, desired_nodes = self._policy.optimize(
                jobs, nodes, prev_allocations, node_template)
            if desired_nodes < len(nodes):
                active_nodes = list(set.union(*map(set, allocations.values())))
            else:
                active_nodes = list(nodes)
                while len(active_nodes) < desired_nodes:
                    active_nodes.append(f"~{desired_nodes-len(active_nodes)}")
            self._cluster_expander.fit(active_nodes)
            LOG.info("Active nodes: %s", active_nodes)
        elif jobs:
            # Expand job ASG from zero nodes.
            # Assumption is AdaptDL is running on a different ASG
            self._cluster_expander.fit(['~1'])
        return allocations


if __name__ == "__main__":
    logging.basicConfig()
    kubernetes.config.load_incluster_config()

    expander = ClusterExpander()
    allocator = AdaptDLAllocator(expander)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(
        expander.run(),
        allocator.run(),
    ))
    loop.close()



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import adaptdl_sched.config as config
import kubernetes_asyncio as kubernetes
import math

from typing import List, Dict, Union


def get_node_unrequested(
        node: kubernetes.client.V1Node,
        pods: List[kubernetes.client.V1Pod],
        ) -> Dict[str, int]:
    """
    Get the amount of node resources which are unrequested by a list of pods.

    Args:
        node: The node to get unrequested resources for.
        pods: Pods which may request resources from the node.

    Returns:
        Mapping from resource names (eg. cpu, memory, nvidia.com/gpu) to
        an integer amount of the resource which is unrequested on the node.
        The integer amounts are discretized to the smallest possible unit of
        each resource.
    """
    ret = {key: _discretize_resource(key, val)
           for key, val in node.status.allocatable.items()}
    for pod in pods:
        if pod.spec.node_name != node.metadata.name \
                or pod.status.phase in ["Succeeded", "Failed"]:
            continue
        requests = get_pod_requests(pod.spec)
        for key, val in requests.items():
            if key in ret:
                ret[key] -= val
    # Note that resources can get negative, as Non-terminated Pods include
    # both Running + Pending pods, and there may be Pending pods yet to be
    # scheduled specifically on each node.
    return {key: val for key, val in ret.items() if val > 0}


def get_pod_requests(
        pod_spec: Union[kubernetes.client.V1PodSpec, dict],
        ) -> Dict[str, int]:
    """
    Get the aggregate amount of resources requested by all containers in a pod.

    Args:
        pod_spec: The pod to get requested resources for.

    Returns:
        Mapping from resource names (eg. cpu, memory, nvidia.com/gpu) to
        an integer amount of the resource which is requested by the pod.
        The integer amounts are discretized to the smallest possible unit of
        each resource.
    """
    # Aggregate resource requests of a pod.
    if not isinstance(pod_spec, dict):
        pod_spec = pod_spec.to_dict()
    overcommitable_resources = ["cpu", "memory", "ephemeral-storage"]
    pod_requests = {"pods": 1}
    for container in pod_spec["containers"]:
        if not container.get("resources"):
            continue
        requests = container["resources"].get("requests") or {}
        for key in overcommitable_resources:
            val = requests.get(key)
            if val is not None:
                pod_requests.setdefault(key, 0)
                pod_requests[key] += _discretize_resource(key, val)
        # For non-overcommitable resources (including extended resources),
        # requests are equal to limits.
        limits = container["resources"].get("limits") or {}
        for key, val in limits.items():
            if key not in overcommitable_resources and val is not None:
                pod_requests.setdefault(key, 0)
                pod_requests[key] += _discretize_resource(key, val)
    return {key: val for key, val in pod_requests.items() if val > 0}


def set_default_resources(pod_spec: dict) -> dict:
    """
    Set the default resources for a given AdaptDLJob spec.

    Args:
        pod_spec: The pod spec to set default resources for.

    Returns:
        A new pod spec with default resources set.
    """
    pod_spec = copy.deepcopy(pod_spec)
    default_resources = config.get_job_default_resources()
    if default_resources is not None:
        # Set default resources for main container.
        container = pod_spec["containers"][0]
        container.setdefault("resources", {})
        if default_resources.get("requests") is not None:
            container["resources"].setdefault("requests", {})
            for k, v in default_resources["requests"].items():
                container["resources"]["requests"].setdefault(k, v)
        if default_resources.get("limits") is not None:
            container["resources"].setdefault("limits", {})
            for k, v in default_resources["limits"].items():
                container["resources"]["limits"].setdefault(k, v)
    return pod_spec


def _discretize_resource(name, value):
    # Normalize to the smallest integral units, try to handle all cases in
    # https://kubernetes.io/docs/concepts/configuration/manage-compute-resources-container
    factor = 1000 if name == "cpu" else 1  # Convert CPU to mCPU.
    # https://github.com/kubernetes/apimachinery/blob/41f3f7bd69163010f79a448955beb2260b91811b/pkg/api/resource/quantity.go#L30
    if isinstance(value, str) and value.endswith("m"):
        factor /= 1000
        value = value[:-1]
    elif isinstance(value, str):
        for idx, unit in enumerate(["k", "M", "G", "T", "P", "E"]):
            if value.endswith(unit):
                factor *= 1000 ** (idx + 1)
                value = value[:-1]
        for idx, unit in enumerate(["Ki", "Mi", "Gi", "Ti", "Pi", "Ei"]):
            if value.endswith(unit):
                factor *= 1024 ** (idx + 1)
                value = value[:-2]
    return math.ceil(float(value) * factor)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import logging
import numpy as np
import pymoo.core.crossover
import pymoo.core.mutation
import pymoo.core.problem
import pymoo.core.repair
import pymoo.optimize

from collections import OrderedDict
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.operators.crossover.util import crossover_mask
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class PolluxPolicy(object):
    def __init__(self):
        self._prev_states = None
        self._prev_jobs = None
        self._prev_nodes = None
        # Utilization thresholds for cluster autoscaling.
        self._min_util = 0.35
        self._max_util = 0.65

    def allocate_job(self, job_info, nodes):
        """
        A simple strategy that find the first available node for a new job.
        This method is intended to allocate a single arriving job. It expects
        the node resources to take into account adaptdl and non-adaptdl pods.

        Arguments:
            job_info (JobInfo): JobInfo object of the job
            nodes (dict): dict from node name to node_info

        Returns:
            list(str): allocation of the job,
                e.g. [node name 0, node name 1, ...] if found available
                     node, else an empty list.
        """
        job_resources = job_info.resources
        min_replicas = max(job_info.min_replicas, 1)
        node_list = []
        nodes = self._sort_nodes(nodes)
        for node_name, node in nodes.items():
            # number of replica fit in this node
            replica_this = min(node.resources.get(key, 0) // val
                               for key, val in job_resources.items())
            if replica_this >= min_replicas:
                node_list = [node_name] * min_replicas
                return node_list
        else:
            return []

    def _sort_nodes(self, nodes):
        return OrderedDict(  # Sort preemptible nodes last.
            sorted(nodes.items(), key=lambda kv: (kv[1].preemptible,
                                                  kv[0])))

    def _allocations_to_state(self, allocations, jobs, nodes):
        jobs_index = {key: idx for idx, key in enumerate(jobs)}
        nodes_index = {key: idx for idx, key in enumerate(nodes)}
        state = np.zeros((len(jobs), len(nodes)), dtype=np.int)
        for job_key, alloc in allocations.items():
            for node_key in (key for key in alloc if key in nodes_index):
                state[jobs_index[job_key], nodes_index[node_key]] += 1
        return state

    def _state_to_allocations(self, state, jobs, nodes):
        allocations = {}
        for job_idx, job_key in enumerate(jobs):
            for node_idx, node_key in enumerate(nodes):
                count = state[job_idx, node_idx]
                allocations.setdefault(job_key, []).extend([node_key] * count)
        return allocations

    def _adapt_prev_states(self, jobs, nodes):
        # Adapt the previously saved optimization states to initialize the
        # current genetic algorithm states.
        shape = (len(self._prev_states), len(jobs), 2 * len(nodes))
        states = np.zeros(shape, dtype=np.int)
        jobs_src = [i for i, key in enumerate(self._prev_jobs) if key in jobs]
        jobs_dst = [i for i, key in enumerate(jobs) if key in self._prev_jobs]
        placeholder = len(self._prev_nodes)  # Next placeholder node to copy.
        # Set allocations for physical (non-placeholder) nodes.
        nodes_index = {key: i for i, key in enumerate(self._prev_nodes)}
        for i, key in enumerate(nodes):
            if key in nodes_index:
                states[:, jobs_dst, i] = \
                    self._prev_states[:, jobs_src, nodes_index[key]]
            elif placeholder < self._prev_states.shape[2]:
                # New node, use allocations for a previous placeholder node.
                states[:, jobs_dst, i] = \
                    self._prev_states[:, jobs_src, placeholder]
                placeholder += 1
        # Set allocations for placeholder nodes.
        for i in range(len(nodes), 2 * len(nodes)):
            if placeholder < self._prev_states.shape[2]:
                states[:, jobs_dst, i] = \
                    self._prev_states[:, jobs_src, placeholder]
                placeholder += 1
        return states

    def _select_result(self, values, max_nodes):
        if np.amin(values[:, 1]) > max_nodes:
            return None
        return np.argmin(np.where(values[:, 1] <= max_nodes, values[:, 0], 0))

    def _desired_nodes(self, utilities, values, nodes):
        idx = self._select_result(values, len(nodes))
        if idx is not None and \
                self._min_util <= utilities[idx] <= self._max_util:
            return len(nodes)
        target_util = (self._min_util + self._max_util) / 2
        best_util = np.inf
        best_nodes = len(nodes)
        for util, (_, num_nodes) in zip(utilities, values):
            if util < self._min_util:
                continue
            if np.isclose(util, best_util) and num_nodes > best_nodes:
                best_nodes = num_nodes
            if abs(util - target_util) < abs(best_util - target_util):
                best_util = util
                best_nodes = num_nodes
        return int(best_nodes)

    def optimize(self, jobs, nodes, base_allocations, node_template):
        """
        Run one optimization cycle of the Pollux scheduling policy.
        This method expects the node resources to only take into account
        non-adaptdl pods.

        Arguments:
            jobs (dict): map from job keys to `JobInfo` objects which
                correspond to the incomplete jobs which should be optimized.
            nodes (dict): map from node keys to `NodeInfo` objects which
                correspond to the existing nodes in the cluster.
            base_allocations (dict): map from job keys to their current
                resource allocations, in the form of a list of a node key for
                each replica.
            node_template (NodeInfo): represents a node which can be requested,
                used to decide the cluster size for cluster auto-scaling.

        Returns:
            dict: map from job keys to their optimized resource allocations,
                in the form of a list of a node key for each replica.
        """

        # A job is considered pinned if it's non-preemptible *and* already has
        # an allocation.
        def ispinned(key, job):
            return not job.preemptible and base_allocations.get(key, []) != []

        # We sort the jobs based on min_replicas and then creation_timestamp,
        # so jobs wanting lower or no min_replicas guarantees are prioritized
        # ahead of those wanting higher min_replicas guarantees to avoid
        # underutilization of cluster. Within a same min_replicas value, they
        # will follow FIFO order. Pinned jobs are aggregated at front because
        # they already have an allocation and won't affect allocations of the
        # rest of the jobs.
        jobs = OrderedDict(sorted(jobs.items(),
                                  key=lambda kv: (not ispinned(kv[0], kv[1]),
                                                  kv[1].min_replicas,
                                                  kv[1].creation_timestamp)))
        nodes = self._sort_nodes(nodes)
        base_state = np.concatenate(
            (self._allocations_to_state(base_allocations, jobs, nodes),
             np.zeros((len(jobs), len(nodes)), dtype=np.int)), axis=1)

        if self._prev_states is None:
            states = np.expand_dims(base_state, 0)
        else:
            states = self._adapt_prev_states(jobs, nodes)
        problem = Problem(list(jobs.values()), list(nodes.values()) +
                          len(nodes) * [node_template], base_state)
        algorithm = NSGA2(
            pop_size=100,
            # pymoo expects a flattened 2-D array.
            sampling=states.reshape(states.shape[0], -1),
            crossover=Crossover(),
            mutation=Mutation(),
            repair=Repair(),
        )
        result = pymoo.optimize.minimize(problem, algorithm, ("n_gen", 100))
        states = result.X.reshape(result.X.shape[0], len(jobs), 2 * len(nodes))
        self._prev_states = copy.deepcopy(states)
        self._prev_jobs = copy.deepcopy(jobs)
        self._prev_nodes = copy.deepcopy(nodes)
        # Get the pareto front.
        nds = NonDominatedSorting().do(result.F, only_non_dominated_front=True)
        states = states[nds]
        values = result.F[nds]
        # Construct return values.
        utilities = problem.get_cluster_utilities(states)
        desired_nodes = self._desired_nodes(utilities, values, nodes)
        idx = self._select_result(values, min(len(nodes), desired_nodes))
        LOG.info("\n" + "-" * 80)
        for i, state in enumerate(states):
            out = "Solution {}:\n".format(i)
            out += "{}\n".format(state)
            out += "Value: {}\n".format(values[i].tolist())
            out += "Utility: {}\n".format(utilities[i])
            out += "-" * 80
            LOG.info(out)
        return (self._state_to_allocations(states[idx], jobs, nodes)
                if idx is not None else {}), desired_nodes


class Problem(pymoo.core.problem.Problem):
    def __init__(self, jobs, nodes, base_state):
        """
        Multi-objective optimization problem used by PolluxPolicy to determine
        resource allocations and desired cluster size. Optimizes for the best
        performing cluster allocation using only the first N nodes. The cluster
        performance and N are the two objectives being optimized, resulting in
        a set of Pareto-optimal solutions.

        The optimization states are a 3-D array of replica assignments with
        shape (pop_size x num_jobs x num_nodes). The element at k, j, n encodes
        the number of job j replicas assigned to node n, in the kth solution.

        Arguments:
            jobs (list): list of JobInfo objects describing the incomplete jobs
                which need to be scheduled.
            nodes (list): list of NodeInfo objects describing the nodes in the
                cluster, in decreasing order of allocation preference.
            base_state (numpy.array): base optimization state corresponding to
                the current cluster allocations. Shape: (num_jobs x num_nodes).
        """
        assert base_state.shape == (len(jobs), len(nodes))
        self._jobs = jobs
        self._nodes = nodes
        self._base_state = base_state
        self._pinned_indices = [i for i, job in enumerate(self._jobs)
                                if not job.preemptible and
                                np.any(self._base_state[i])]
        # Find which resource types are requested by at least one job.
        rtypes = sorted(set.union(*[set(job.resources) for job in jobs]))
        # Build array of job resources: <num_jobs> x <num_rtypes>. Each entry
        # [j, r] is the amount of resource r requested by a replica of job j.
        self._job_resources = np.zeros((len(jobs), len(rtypes)), np.int64)
        for j, job in enumerate(jobs):
            for r, rtype in enumerate(rtypes):
                self._job_resources[j, r] = job.resources.get(rtype, 0)
        # Build array of node resources: <num_nodes> x <num_rtypes>. Each
        # entry [n, r] is the amount of resource r available on node n.
        self._node_resources = np.zeros((len(nodes), len(rtypes)), np.int64)
        for n, node in enumerate(nodes):
            for r, rtype in enumerate(rtypes):
                self._node_resources[n, r] = node.resources.get(rtype, 0)
        # Calculate dominant per-replica resource shares for each job.
        shares = self._job_resources / np.sum(self._node_resources, axis=0)
        self._dominant_share = np.amax(shares, axis=1)
        # Upper bound each job: <replicas on node 0> <replicas on node 1> ...
        self._max_replicas = np.zeros(base_state.shape, dtype=np.int)
        for j, job in enumerate(jobs):
            for n, node in enumerate(nodes):
                self._max_replicas[j, n] = min(
                    self._get_avail_resource(
                        n, node, rtype) // job.resources[rtype]
                    for rtype in rtypes if job.resources.get(rtype, 0) > 0)
        self._restart_penalty = 0.1
        # Lower bound each job by min_replicas from job spec
        self._min_replicas = np.zeros(base_state.shape, dtype=np.int)
        for j, job in enumerate(jobs):
            min_replicas = self._jobs[j].min_replicas
            for n, node in enumerate(nodes):
                self._min_replicas[j, n] = min(min_replicas,
                                               self._max_replicas[j, n])
                min_replicas -= self._min_replicas[j, n]
        super().__init__(n_var=self._base_state.size, n_obj=2, type_var=np.int)

    def _get_avail_resource(self, node_idx, node, rtype):
        # Cutoff node's maximum allowable resources by amount already used by
        # pinned jobs.
        resource = node.resources.get(rtype, 0)
        allocs = self._base_state[self._pinned_indices]
        for job_idx, alloc in enumerate(allocs):
            resource -= alloc[node_idx] * \
                self._jobs[self._pinned_indices[job_idx]] \
                .resources.get(rtype, 0)
        assert resource >= 0
        return resource

    def get_cluster_utilities(self, states):
        """
        Calculates the cluster utility for each state, defined as the average
        percentage of ideal speedup for each job (ie. speedup / num_replicas),
        weighted by the job's share of the most congested cluster resource.

        Arguments:
            states (numpy.array): a (pop_size x num_jobs x num_nodes) array
                containing the assignments of job replicas to nodes.

        Returns:
            numpy.array: a (pop_size) array containing the utility for each
                state.
        """
        num_replicas = np.sum(states, axis=2)
        speedups = self._get_job_speedups(states)
        # mask (pop_size x num_nodes): indicates which nodes are active.
        mask = np.sum(states, axis=1) > 0
        # total (pop_size x num_rtypes): total amount of cluster resources.
        total = np.sum(np.expand_dims(mask, 2) * self._node_resources, axis=1)
        # alloc (pop_size x num_jobs x num_rtypes):
        #     amount of cluster resources allocated to each job.
        alloc = np.expand_dims(num_replicas, 2) * self._job_resources
        with np.errstate(divide="ignore", invalid="ignore"):
            # shares (pop_size x num_jobs x num_rtypes):
            #     resource shares for each job as a fraction of the cluster.
            shares = np.where(alloc, alloc / np.expand_dims(total, 1), 0.0)
            # utilities (pop_size x num_jobs):
            #     utilities for each job as a fraction of ideal scalability.
            utilities = np.where(num_replicas, speedups / num_replicas, 0.0)
        # Weighted average across all jobs for each rtype.
        utilities = np.sum(np.expand_dims(utilities, 2) * shares, axis=1)
        # Return the utilities for the best utilized rtypes.
        return np.amax(utilities, axis=1)  # Shape: (pop_size).

    def _get_job_speedups(self, states):
        speedup = []
        num_nodes = np.count_nonzero(states, axis=2)
        num_replicas = np.sum(states, axis=2)
        for idx, job in enumerate(self._jobs):
            speedup.append(job.speedup_fn(
                num_nodes[:, idx], num_replicas[:, idx]))
        return np.stack(speedup, axis=1).astype(np.float)

    def _get_cluster_sizes(self, states):
        sizes = np.arange(len(self._nodes)) + 1
        return np.amax(np.where(np.any(states, axis=-2), sizes, 0), axis=-1)

    def _evaluate(self, states, out, *args, **kwargs):
        states = states.reshape(states.shape[0], *self._base_state.shape)
        speedups = self._get_job_speedups(states)
        # Scale the speedup of each job so that a dominant resource share
        # equivalent to a single node results in a speedup of 1.
        scaled_speedups = speedups * self._dominant_share * len(self._nodes)
        # Penalize job restarts.
        restart_mask = np.any(states != self._base_state, axis=2)
        scaled_speedups[restart_mask] *= 1.0 - self._restart_penalty
        out["F"] = np.column_stack([-np.sum(scaled_speedups, axis=1),
                                    self._get_cluster_sizes(states)])

    def _crossover(self, states, **kwargs):
        states = states.reshape(*states.shape[:2], *self._base_state.shape)
        n_parents, n_matings, n_jobs, n_nodes = states.shape
        # Single-point crossover over jobs for all parent states.
        points = np.random.randint(n_jobs, size=(n_matings, 1))
        result = crossover_mask(states, np.arange(n_jobs) < points)
        # Set cluster sizes uniformly at random between each pair of parents.
        min_nodes, max_nodes = np.sort(self._get_cluster_sizes(states), axis=0)
        num_nodes = np.random.randint(np.iinfo(np.int16).max,
                                      size=(n_parents, n_matings))
        num_nodes = min_nodes + num_nodes % (max_nodes - min_nodes + 1)
        mask = np.arange(n_nodes) >= np.expand_dims(num_nodes, (2, 3))
        result[np.broadcast_to(mask, result.shape)] = 0
        return result.reshape(n_parents, n_matings, -1)

    def _mutation(self, states, **kwargs):
        # Select variables randomly, then assign each of them a random value
        # within their upper bounds.
        states = states.reshape(states.shape[0], *self._base_state.shape)
        num_nonzero = np.count_nonzero(states, axis=2, keepdims=True)
        num_zero = states.shape[2] - num_nonzero
        # Try to balance the number of mutations between zero/nonzero elements.
        prob = 1.0 / np.where(states > 0, num_nonzero, num_zero)
        prob = prob.reshape(states.shape)
        m = np.random.random(states.shape) < prob
        r = np.random.randint(self._min_replicas, self._max_replicas + 1,
                              size=states.shape)
        states[m] = r[m]
        # We need at least min_replicas
        states = np.maximum(states, self._min_replicas)
        return states.reshape(states.shape[0], -1)

    def _repair(self, pop, **kwargs):
        states = pop.get("X")
        states = states.reshape(states.shape[0], *self._base_state.shape)
        # Copy previous allocations for pinned jobs
        states[:, self._pinned_indices] = \
            self._base_state[self._pinned_indices, :]
        # Enforce at most one distributed job per node. Exclude all
        # nonpreemptible jobs.
        distributed = np.count_nonzero(states, axis=2) > 1
        mask = states * np.expand_dims(distributed, axis=-1) > 0
        mask = mask.cumsum(axis=1) > 1
        states[mask] = 0
        # Enforce no more than max replicas per job.
        # max_replicas: (num_jobs x 1)
        max_replicas = np.array([[j.max_replicas] for j in self._jobs])
        shuffle = np.argsort(np.random.random(states.shape), axis=2)
        states = np.take_along_axis(states, shuffle, axis=2)  # Shuffle nodes.
        states = np.minimum(np.cumsum(states, axis=2), max_replicas)
        states = np.diff(states, axis=2, prepend=0)
        inverse = np.argsort(shuffle, axis=2)  # Undo shuffle nodes.
        states = np.take_along_axis(states, inverse, axis=2)
        # Enforce node resource limits.
        # job_resources: (num_jobs x num_nodes x num_rtypes)
        job_resources = np.expand_dims(self._job_resources, 1)
        states = np.expand_dims(states, -1) * job_resources
        states = np.minimum(np.cumsum(states, axis=1), self._node_resources)
        states = np.diff(states, axis=1, prepend=0)
        with np.errstate(divide="ignore", invalid="ignore"):
            states = np.amin(np.floor_divide(states, job_resources),
                             where=job_resources > 0, initial=99, axis=-1)
        # Only choose solutions which have at least min_replicas allocations
        min_replicas = np.array([j.min_replicas for j in self._jobs])
        mask = np.sum(states, axis=-1) < min_replicas
        states[mask] = 0
        return pop.new("X", states.reshape(states.shape[0], -1))


class Crossover(pymoo.core.crossover.Crossover):
    def __init__(self):
        super().__init__(n_parents=2, n_offsprings=2)

    def _do(self, problem, states, **kwargs):
        return problem._crossover(states, **kwargs)


class Mutation(pymoo.core.mutation.Mutation):
    def _do(self, problem, states, **kwargs):
        return problem._mutation(states, **kwargs)


class Repair(pymoo.core.repair.Repair):
    def _do(self, problem, pop, **kwargs):
        return problem._repair(pop, **kwargs)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os

ADAPTDL_PH_LABEL = 'adaptdl/placeholder'


def allowed_taints(taints):
    if not taints:
        return True
    return (len(taints) == 1 and taints[0].key == "petuum.com/nodegroup" and
            taints[0].value == "adaptdl")


def get_namespace():
    # for code running outside of AdaptDL
    if not os.path.exists("/var/run/secrets/kubernetes.io/serviceaccount/namespace"):  # noqa: E501
        return "default"
    with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace") as f:
        return f.read()


def get_image():
    return os.environ["ADAPTDL_IMAGE"]


def get_adaptdl_deployment():
    return os.environ["ADAPTDL_SCHED_DEPLOYMENT"]


def get_supervisor_url():
    return os.environ["ADAPTDL_SUPERVISOR_URL"]


def get_supervisor_port():
    return os.getenv("ADAPTDL_SUPERVISOR_SERVICE_PORT", 8080)


def get_storage_subpath():
    return os.environ["ADAPTDL_STORAGE_SUBPATH"]


def get_adaptdl_version():
    return os.environ["ADAPTDL_SCHED_VERSION"]


def get_job_default_resources():
    val = os.getenv("ADAPTDL_JOB_DEFAULT_RESOURCES")
    return json.loads(val) if val is not None else None


def get_job_patch_pods():
    val = os.getenv("ADAPTDL_JOB_PATCH_PODS")
    return json.loads(val) if val is not None else None


def get_job_patch_containers():
    val = os.getenv("ADAPTDL_JOB_PATCH_CONTAINERS")
    return json.loads(val) if val is not None else None
